.lutris-logview > text {
  color: rgb(255, 199, 116);
  background-color: rgb(47, 47, 47);
}

.lutris-logview > text > selection {
  color: rgb(47, 47, 47);
  background-color: rgb(255, 199, 116);
}

.lutris-sidebar button {
  padding: 0;
  border: 0;
}

.lutris-sidebar button > image {
  padding: 0;
  border: 0;
}

.info-pill {
  background-color: alpha(currentColor, .07);
  border-radius: 5px;
  padding: 3px 6px;
}

.option-wrapper-assigned-here > label {
    font-weight: bold;
}

.option-wrapper-non-default > label {
    font-style: italic;
}

.section-frame {
	margin-top: 12px;
	margin-bottom: 12px;
	margin-left: 12px;
	margin-right: 12px;
}

.section-frame > box {
	margin-bottom: 6px;
}

.reset-button {
    padding: 0px;
}

.info-frame {
  color: @theme_text_color;
  background-color: @theme_base_color;
}

.in-app-notification {
    padding: 5px;
    border-radius: 0
}

.in-app-notification label link {
    color: unset;
}

.in-app-notification button {
	padding: 0px;
	min-height: 0px;
	min-width: 0px;
	border-radius: 0px;
}

.smallcheckbox label {
   font-size: 10px
}

.smallcheckbox check {
  min-width: 8px;
  min-height: 8px;
  border-bottom-left-radius: 2px;
  border-bottom-right-radius: 2px;
  border-top-left-radius: 2px;
  border-top-right-radius: 2px;
  margin-right: 2px
}

.cuddledcheckbox {
  padding: 0px
}

.cuddledcheckbox check {
  margin-left: 0px;
}

.cuddledcheckbox label {
  margin-right: 0px
}